package cz.jikos.ales.mff.prgj.cfinder;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.text.ParseException;
import java.util.LinkedList;

public class Indexer extends ClassFinder {

	protected Storage storageOld;
	protected Parser parser;

	protected int maxCount = 0;
	protected int count = 0;
	protected int countCached = 0;
	protected boolean printMatchedFiles = false;
	protected boolean followSymlinks = false;
	
	protected static long logtime = 0;
	protected static boolean isDebug;
	
	public static void log (String s) {
		
		long nt, tim;
		
		if (! isDebug) return;
		tim = System.currentTimeMillis();
		
		nt = tim - logtime;
		logtime = tim;
		System.out.printf ("[%010d] %s\n", nt, s);
		System.out.flush();
	}
	
	public Indexer (Storage storage, Storage storageOld, Parser parser) 
	{
		this.storage = storage;
		this.storageOld = storageOld;
		this.parser = parser;
		this.dirsTodo = new LinkedList<File>();
	}
	
	public void addPath (String path)
	{
		dirsTodo.add (new File (path));
	}
	
	public void announceDirectory (String path) {
		//System.out.println ("dir: " + path);
	}
	
	public void announceFile (String path) {
		if (printMatchedFiles)
			System.out.println ("classfile: " + path);
	}
	
	/**
	 * Indexes one file.
	 * 
	 * If storageOld is specified, it is first searched for specified path.
	 * If found, saved record is retrieved and if its lastModified time 
	 * is equals to current lastModified time of file, it is not parsed again.
	 * 
	 * @param f
	 * @param path
	 * @return
	 */
	protected boolean indexClassfile (File f, String path) {
		try {
			ClsInfo clsInfo;
			if (storageOld != null) {
				ClsInfo query = new ClsInfo();
				query.fileName = path;
				
				ClsInfo [] cs = storageOld.findClsInfo(query, 1, null);
				if (cs.length == 1 && cs [0].date == f.lastModified()) {
					clsInfo = cs [0];
					clsInfo.date = f.lastModified();
					storage.storeClsInfo(path, clsInfo);
					countCached ++;
					return true;
				}
			} 
			clsInfo = parser.parse(new FileInputStream(f));
			clsInfo.date = f.lastModified();

			log ("Before store");
			storage.storeClsInfo(path, clsInfo);
			log ("After store");
			return true;
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			return false;
		} catch (ParseException e) {
			return false;			
		}
	}
	/**
	 * Routine traverses filesystem.
	 * On each *.class file found, it calls @ref indexClassfile
	 */
	public void index ()
	{
		
		while (! dirsTodo.isEmpty()) {
			
			File dir = dirsTodo.pop();
			
			announceDirectory(dir.getAbsolutePath());
			
			File [] files = dir.listFiles();
			
			if (files == null) continue;
			for (File f : files) {
				if (f.isDirectory()) {
					
					try {
						/* this trick eliminates symbolic links ... if we are lucky */
						if (followSymlinks || f.getCanonicalPath().equals(f.getAbsolutePath()))

							dirsTodo.add(f);
					} catch (IOException e) {
						// we just miss that dir
					}
				} else if (f.isFile()) {
					String path = f.getAbsolutePath();
					if (path.endsWith(".class")) {
						announceFile (path);
						
						if (indexClassfile (f, path)) {
							if ((maxCount > 0) && (++count > maxCount)) return;
						}
					}
				}
			}
		}
	}
	/**
	 * Create indexer using commandline arguments.
	 * 
	 * @param args
	 */
	public Indexer (String []args) {
		parseArgs(args);
		initializePaths();
		initializeStorage();
		parser = new Parser();

		if (dirsTodo.isEmpty()) addPath(home);
		
		if (hasToIndex) {
			if (! storage.isEmpty()) {
				/*
				 * if we have to index, and there is (nonepty) index
				 * already, we:
				 *  - create new index in subdirectory "_tmp" of current
				 *    index
				 *  - proceed into new index, using old one
				 *  - when we're done, rename new index to old
				 */
				String storageDirSave = storageDir, storageDirTmp = storageDir + File.separator + "_tmp";
				storageDir = storageDirTmp;
				storageOld = storage;
				storage = null;
				initializeStorage();
				index();
				storageOld.close();
				storage.move(storageDirTmp, storageDirSave);
				storage.close();
				storageDir = storageDirSave;
				initializeStorage();
			} else {
				index();
			}
		} else {
			if (hasToQuery) Query(System.out);
		}
	}
	
	/**
	 * Print ClsInfo to stream.
	 * 
	 * @param cinfo
	 * @param ps
	 */
	public void printClsInfo (ClsInfo cinfo, PrintStream ps)
	{
		ps.printf("%s.%s\n file:%s\n", (cinfo.pkg!=null)?cinfo.pkg.name:"", cinfo.name, cinfo.fileName);
		ps.printf(" methods:\n");
		for (ClsInfo.Method m : cinfo.methods) {
			ps.printf ("  %s\n", m.name);
		}
		for (ClsInfo.Field f : cinfo.fields) {
			ps.printf ("  %s\n", f.name);
		}
	}
	
	/**
	 * Do Query.
	 * 
	 * Create ClsInfo query object and look into storage.
	 * Print found objects to OutputStream.
	 * 
	 * @param os
	 */
	public void Query (OutputStream os) {
		ClsInfo query = new ClsInfo ();
		final PrintStream ps = new PrintStream (os);
		if (queryClassName != null) query.name = queryClassName;
		if (queryFileName != null) query.fileName = queryFileName;
		if (queryPackageName != null) query.pkg = new PkgInfo(queryPackageName);
		
		if (queryMethodName != null) {
			ClsInfo.Method m = new ClsInfo.Method ();
			m.name = queryMethodName;
			query.methods.add(m);
		}
		
		
		Storage.Callback callback = new Storage.Callback () {
			@Override
			public boolean call(ClsInfo object) {
				printClsInfo (object, ps);
				
				return false;
			}
		};
		storage.findClsInfo(query, 500, callback);
	}	
}
