package cz.jikos.ales.mff.prgj.cfinder;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.text.ParseException;
import java.util.concurrent.locks.ReadWriteLock;


/*
 * 1	2+x bytes
(variable)	UTF-8 (Unicode) string: a character string prefixed by a 16-bit number (type u2) indicating the number of bytes in the encoded string which immediately follows (which may be different than the number of characters). Note that the encoding used is not actually UTF-8, but involves a slight modification of the Unicode standard encoding form.
3	4 bytes	Integer: a signed 32-bit two's complement number in big-endian format
4	4 bytes	Float: a 32-bit single-precision IEEE 754 floating-point number
5	8 bytes	Long: a signed 64-bit two's complement number in big-endian format (takes two slots in the constant pool table)
6	8 bytes	Double: a 64-bit double-precision IEEE 754 floating-point number (takes two slots in the constant pool table)
7	2 bytes	Class reference: an index within the constant pool to a UTF-8 string containing the fully qualified class name (in internal format)
8	2 bytes	String reference: an index within the constant pool to a UTF-8 string
9	4 bytes	Field reference: two indexes within the constant pool, the first pointing to a Class reference, the second to a Name and Type descriptor.
10	4 bytes	Method reference: two indexes within the constant pool, the first pointing to a Class reference, the second to a Name and Type descriptor.
11	4 bytes	Interface method reference: two indexes within the constant pool, the first pointing to a Class reference, the second to a Name and Type descriptor.
12	4 bytes	Name and type descriptor: two indexes to UTF-8 strings within the constant pool, the first representing a name (identifier) and the second a specially encoded type descriptor.
 */
/**
 * Incomporates simple Java classfile parser.
 * 
 * @author alesak@jikos.cz
 */
public class Parser {

	/**
	 * All the constants are stored in constants table in the beginning 
	 * of the classfile.
	 * 
	 * @author alesak@jikos.cz
	 */
	protected static class ConstRecord {
		public enum ConstType {
			UtfString(1),
			Integer(3),
			Float(4),
			Long(5),
			Double(6),
			Class(7),
			String(8),
			Field(9),
			Method(10),
			Interface(11),
			Name(12);
			
			private int id;
			static ConstType[] table = null;
			
			public static ConstType byId(int id) throws ParseException{
				if (table == null) {
					int max = 0;
					for (ConstType t : values()) 
						if (t.id > max) max = t.id;
					
					table = new ConstType[max + 1];
					
					for (ConstType t : values())
						table [t.id] = t;
				}
				if (table[id] == null) {
					throw new ParseException("Unknown const type " + id, 0);
				}
				return (table [id]);
			}
			
			ConstType (int id) {
				this.id = id;
			}
			
		}
		ConstType type;
		
		protected String stringValue;
		protected long longValue;
		protected int intValue;
		
		protected ConstRecord ref;
		protected int refId = -1;
		
		public String toString()
		{
			return super.toString() + ":" + type;				
		}
		
		public ConstRecord (InputStream istream) throws IOException, ParseException
		{
			byte []b = new byte[8];
			int id, i2;
			int tid = istream.read();
			
			type = ConstType.byId(tid);
			if (type == null) 
			throw new RuntimeException("Unknown const type");
			switch (type) {
			case UtfString:
				int len = Parser.readWord(istream);
				stringValue = Parser.readString (istream, len);
				break;
			case Integer:
				Parser.readBuff(istream, b, 4);
				break;
			case Float:
				Parser.readBuff(istream, b, 4);
				break;
			case Long:
				Parser.readBuff(istream, b, 8);
				break;
			case Double:
				Parser.readBuff(istream, b, 8);
				break;
			case Class:
				refId = Parser.readWord(istream);
				break;
			case String:
				id = Parser.readWord(istream);
				break;
			case Field:
				id = Parser.readWord(istream);
				i2 = Parser.readWord(istream);
				break;
			case Method:
				id = Parser.readWord(istream);
				i2 = Parser.readWord(istream);
				break;
			case Interface:
				id = Parser.readWord(istream);
				i2 = Parser.readWord(istream);
				break;
			case Name:
				id = Parser.readWord(istream);
				i2 = Parser.readWord(istream);
				break;
			default:
				throw new ParseException("Unknown type", 0);
			}
			//System.out.printf ("Const %s %d\n", type.toString(), tid);
		}
	}
	
	/**
	 * Describing entity attributes (wildly used in Java classfile
	 * to specify anything, e.g. bytecode position).
	 *  
	 * @author alesak@jikos.cz
	 */
	protected static class AttrRecord {
		ConstRecord name;
		
		public AttrRecord (InputStream istream, ConstRecord []consts) throws IOException, ParseException
		{
			int idname = Parser.readWord (istream);
			if (idname > 0) name = consts [idname - 1];
			int cnt = Parser.readInt(istream);
			//System.out.printf ("idname=%d,name=%s,len=%d\n", idname, name.stringValue, cnt);
			if (cnt == -1) return;
			byte []b = new byte[cnt];
			Parser.readBuff(istream, b, cnt);
		}
	}
	
	/**
	 * Describing class fields.
	 * 
	 * @author alesak@jikos.cz
	 */
	protected static class FieldRecord {
		int access;
		
		ConstRecord name;
		ConstRecord desc;
		AttrRecord []attrs;
		
		public FieldRecord (InputStream istream, ConstRecord []consts) throws IOException, ParseException
		{
			access = Parser.readWord (istream);
			name = consts [Parser.readWord (istream) - 1];
			desc = consts [Parser.readWord (istream) - 1];
			int cnt = Parser.readWord(istream);
			//System.out.printf("FieldRecord.Name %s \"%s\", attrcnt=%d\n", name, name.stringValue, cnt);
			attrs = new AttrRecord[cnt];
			for (int i = 0; i < cnt; i ++) {
				attrs [i] = new AttrRecord (istream, consts);
			}
		}
	}
	
	/**
	 * Structure describing class method.
	 * 
	 * @author alesak@jikos.cz
	 */
	protected static class MethodRecord {
		int access;
		ConstRecord name;
		ConstRecord desc;
		AttrRecord []attrs;
		
		public MethodRecord (InputStream istream, ConstRecord []consts) throws IOException, ParseException
		{
			access = Parser.readWord (istream);
			name = consts [Parser.readWord (istream) - 1];
			//System.out.printf("Name %s \"%s\"\n", name, name.stringValue);
			desc = consts [Parser.readWord (istream) - 1];
			int cnt = Parser.readWord(istream);
			//System.out.printf("MethodRecord.Name %s \"%s\" desc=\'%s\', attrcnt=%d\n", name, name.stringValue, desc.stringValue, cnt);
			attrs = new AttrRecord[cnt];
			for (int i = 0; i < cnt; i ++) {
				
				attrs [i] = new AttrRecord (istream, consts);
				//System.out.printf ("_attr[%d]\n", i);
			}
		}
	}
	
	protected static int readWord (InputStream istream) throws IOException
	{
		return (int)((int)istream.read() << 8  | (int)istream.read()); 
	}
	
	protected static void readBuff (InputStream istream, byte []b, int l) throws ParseException, IOException
	{
		for (int i = 0; i < l; i ++) {
			b [i] = (byte) istream.read();
		}
	}
	protected static String readString (InputStream istream, int len) throws ParseException, IOException
	{
		byte []b = new byte[len];
		if (istream.read(b) != len) throw new ParseException("Cannot read " + len, 0);
		
		return new String(b, "UTF-8");
	}
	
	protected void log(){}
	
	protected static int readInt (InputStream istream) throws IOException {
		return readWord(istream) << 16 | readWord (istream);
	}
	
	protected void debug (String s)
	{
	//	System.out.println (s);
	}
	
	protected void log (String s)
	{
	//	System.out.println (s);
	}
	
	/**
	 * just to test diacritic */
	public void heč () {}
	
	/**
	 * Splits classname into package and classname, switches / for .
	 * @param name
	 * @param cinfo
	 */
	protected void parseClassName (String name, ClsInfo cinfo) {
		name = name.replace('/', '.');
		int pos = name.lastIndexOf('.');
		if (pos != -1) {
			PkgInfo pkgInfo = new PkgInfo (name.substring (0, pos));
			cinfo.pkg = pkgInfo;
			cinfo.name = name.substring(pos + 1);
		} else {
			cinfo.name = name;
		}
	}
	/**
	 * Parses input stream and returns contained classfile information.
	 * 
	 * @param istream
	 * @return
	 * @throws ParseException
	 */
	@SuppressWarnings("unused")
	public ClsInfo parse(InputStream istream) throws ParseException 
	{
		ClsInfo cin = new ClsInfo();
		byte head[] = new byte[4];
		try {
			if (istream.read(head, 0, 4) != 4) 
				throw new ParseException ("Cannot read header", 0);
				
			if (head [0] != (byte)0xCA || head [1] != (byte)0xFE || head [2] != (byte)0xBA || head [3] != (byte)0xBE) throw new ParseException("Not a Java Class", 0);
	
			int versMinor = readWord (istream);
			int versMajor = readWord (istream);
			int constSize = readWord (istream);
			
			constSize --;
			log ("constSize=" + constSize);

			int drift = 0;
			ConstRecord []crec = new ConstRecord[constSize];
			
			for (int i = 0; i < constSize; i ++) {
				crec [i] = new ConstRecord (istream);
				debug ("crec[" + i + "]=" + crec.toString());
				if (crec [i].type == ConstRecord.ConstType.Double || crec [i].type == ConstRecord.ConstType.Long) {
					//constSize --;
					drift ++;
					i ++;
				}
			}
			
			for (int i = 0; i < constSize; i ++) 
				if (crec [i] != null && crec [i].refId != -1)
					crec [i].ref = crec [crec [i].refId - 1];
				
			int accessFlags = readWord (istream);
			int thisId = readWord (istream);
			int superId = readWord (istream);
			
			//cin.name = crec [thisId-1].ref.stringValue;
			parseClassName (crec [thisId-1].ref.stringValue, cin);
			log ("class_name=" + cin.name);
			
			int cnt = readWord (istream);
			//ConstRecord []ifaces = new ConstRecord[cnt];
			int []iifaces = new int[cnt];
			
			for (int i = 0; i < cnt; i ++)
				iifaces [i] = readWord(istream);
			
			cnt = readWord (istream);
			log ("field_cnt=" + cnt);
			
			FieldRecord []frecs = new FieldRecord[cnt];
			
			for (int i = 0; i < cnt; i ++)
				frecs [i] = new FieldRecord(istream, crec);
			
			cnt = readWord (istream);
			log ("method_cnt=" + cnt);
			
			MethodRecord []mrecs = new MethodRecord[cnt];
			
			for (int i = 0; i < cnt; i ++)
				mrecs [i] = new MethodRecord(istream, crec);
			
			for (FieldRecord f : frecs) {
				ClsInfo.Field fi = new ClsInfo.Field();
				fi.name = f.name.stringValue;
				cin.fields.add(fi);
			}
			
			for (MethodRecord m : mrecs) {
				ClsInfo.Method mi = new ClsInfo.Method();
				mi.name = m.name.stringValue;
				cin.methods.add(mi);
			}
			
		} catch (IOException e) {
			e.printStackTrace();
			throw new RuntimeException ("Cannot parse (IO)");
		} catch (ParseException e) {
			e.printStackTrace();
			throw e;
		}
		
		return (cin);
	}

}
