package cz.jikos.ales.mff.prgj.cfinder;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.LinkedList;
import java.util.List;

import cz.jikos.ales.mff.prgj.simpleser.SimpleSer;

/**
 * RandomAccessFile storage, using @ref RafTrie for indexing 
 * key fields.
 * 
 * @author alesak@jikos.cz
 */
public class RafStorage implements Storage {

	protected SimpleSer<ClsInfo> serializer = new SimpleSer<ClsInfo>();
	/** raf-file for class.data */
	protected RandomAccessFile strgMain;
	/** trie for filename.idx */
	protected RafTrie idxFilename;
	/** trie for class.idx */
	protected RafTrie idxClassname;
	/** trie for pkgs.idx */
	protected RafTrie idxPkgname;
	/** trie for method.idx */
	protected RafTrie idxMethodname;

	public RafStorage(File f) {
		if (!f.isDirectory()) {
			throw new RuntimeException(f.getName()
					+ " assumed to be a directory!");
		}

		try {
			File nf = new File(f.getAbsoluteFile() + File.separator
					+ "class.data");
			strgMain = new RandomAccessFile(nf, "rw");
			idxFilename = new RafTrie(new File(f.getAbsoluteFile()
					+ File.separator + "filename.idx"));
			idxClassname = new RafTrie(new File(f.getAbsoluteFile()
					+ File.separator + "class.idx"));
			idxPkgname = new RafTrie(new File(f.getAbsoluteFile()
					+ File.separator + "pkgs.idx"));
			idxMethodname = new RafTrie(new File(f.getAbsoluteFile()
					+ File.separator + "method.idx"));
		} catch (FileNotFoundException e) {
			throw new StorageException("Cannot initialize storage "
					+ e.getMessage());
		} catch (IOException e) {
			e.printStackTrace();
			throw new StorageException("Cannot use storage.");
		}
	}

	/**
	 * Takes key attributes of specified clsInfo along with path
	 * to its location and put them to indexes, pointing at
	 * specified id (which in fact leads to class.data).
	 * 
	 * @param clsInfo
	 * @param path
	 * @param id
	 */
	protected void makeIndexes(ClsInfo clsInfo, String path, long id) {
		idxClassname.putObject(clsInfo.name, id);
		idxFilename.putObject(path, id);
		for (ClsInfo.Method m : clsInfo.methods) {
			idxMethodname.putObject(m.name, id);
		}
		if (clsInfo.pkg != null)
			idxPkgname.putObject(clsInfo.pkg.name, id);
	}

	@Override
	/**
	 * Stores @ref ClsInfo into database, including indexing 
	 * its key fields.
	 * 
	 * @param path path to classfile
	 * @param clsInfo ClsInfo to store
	 */
	public void storeClsInfo(String path, ClsInfo clsInfo) {
		clsInfo.fileName = path;

		long fp;
		try {
			fp = strgMain.getFilePointer();
			FileOutputStream fos = new FileOutputStream(strgMain.getFD());

			try {
				serializer.serialize(clsInfo, fos);

				makeIndexes(clsInfo, path, fp);
			} catch (Exception e) {
				e.printStackTrace();
				fos.flush();
				strgMain.setLength(fp);
			}
		} catch (IOException e) {
			e.printStackTrace();
			throw new StorageException("Cannot store");
		}
	}

	/**
	 * From main storage (class.data) retrieves object starting at
	 * specified position.
	 * 
	 * @param id position where to start
	 * @return object or null, if error occured
	 */
	protected ClsInfo getClsInfoById(long id) {
		try {
			FileInputStream fis = new FileInputStream(strgMain.getFD());
			strgMain.seek(id);
			ClsInfo clsInfo = serializer.deserialize(fis, ClsInfo.class);
			return (clsInfo);
		} catch (Exception e) {
			e.printStackTrace();
			System.out.println("Cannot get object from pos " + id);
			return (null);
		}
	}

	@Override
	public ClsInfo[] findClsInfo(ClsInfo query, int limit, Callback callback) {
		List<ClsInfo> clsInfos = new LinkedList<ClsInfo>();
		int longest = 0;
		RafTrie idx = null;
		String queryString = null;

		if (query.name != null && query.name.length() > longest) {
			idx = idxClassname;
			queryString = query.name;
			longest = queryString.length();
		}
		if (query.pkg != null && query.pkg.name != null
				&& query.pkg.name.length() > longest) {
			idx = idxPkgname;
			queryString = query.pkg.name;
			longest = queryString.length();
		}
		if (query.fileName != null && query.fileName.length() > longest) {
			idx = idxFilename;
			queryString = query.fileName;
			longest = queryString.length();
		}
		for (ClsInfo.Method m : query.methods) {
			if (m.name.length() > longest) {
				idx = idxMethodname;
				queryString = m.name;
				longest = queryString.length();
			}

		}

		if (idx == null) {
			throw new RuntimeException("Cannot determine what to search");
		}

		RafTrie.Callback cb = new RafTrie.Callback() {

			ClsInfo query;

			Callback callback;
			List<ClsInfo> clsInfos;
			int limit;
			int r = 0;

			public RafTrie.Callback setObjects(List<ClsInfo> clsInfos,
					ClsInfo query, int limit, Callback callback) {
				this.clsInfos = clsInfos;
				this.query = query;
				this.limit = limit;
				this.callback = callback;
				return this;
			}

			@Override
			public boolean call(long objId) {
				ClsInfo clsInfo = getClsInfoById(objId);
				if (clsInfo == null)
					return false;
				if (!clsInfo.matches(query, false)) {
					return false;
				} else {
				}
				clsInfos.add(clsInfo);
				if (callback != null) {
					if (callback.call(clsInfo)) {
						return false;
					}
				}
				return (++r >= limit);
			}
		}.setObjects(clsInfos, query, limit, callback);

		if (queryString.endsWith("*")) {
			queryString = queryString.substring(0, queryString.length() - 1);
			idx.findObjectByPrefix(queryString, cb);
		} else
			idx.findObject(queryString, cb);

		return (clsInfos.toArray(new ClsInfo[0]));
	}
	
	@Override
	public void close()
	{
		try {
			strgMain.close ();
		} catch (IOException e) {
			throw new StorageException("Cannot close file...");
		}
		idxClassname.close();
		idxFilename.close();
		idxMethodname.close();
		idxPkgname.close();
	}

	@Override
	public boolean isEmpty() {
		try {
			return (strgMain.length() == 0);
		} catch (IOException e) {
			throw new StorageException("Cannot get file size" + e.getMessage());
		}
	}

	@Override
	public void move(String source, String dest) {
		String files[] = {"class.data", "class.idx", "filename.idx", "method.idx", "pkgs.idx"};
		
		for (String s: files) {
			File sf = new File (source + File.separator + s);
			File df = new File (dest + File.separator + s);
			sf.renameTo(df);
		}
	}
}
