package cz.jikos.ales.mff.prgj.cfinder;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.io.UnsupportedEncodingException;

/**
 * Simple Trie structure over RandomAccessFile.
 * 
 * File is separated in constant (8) size blocks.
 * 
 * 
 * Trie node is formed by continous series of (following) blocks,
 * attached together by single direction links.
 *
 * The last block in node is determined by tag.
 *
 * Node in the first block is taken as root node.
 * 
 * Node can have multiple values (longs).
 * 
 * Each block has this structure:
 * 
 * <pre>
 * +----+----+----+----+----+----+----+----+
 * +    +    +    +    +    +    +    +    +
 * +tag +spec+ b0 + b1 + b2 + b3 +    +    +
 * +    +    +    +    +    +    +    +    +
 * +----+----+----+----+----+----+----+----+
 * 
 * tag is:
 *   z - (zarazka) this is last block in node
 *   n - (next - link) next block has address b3b2b1b0
 *   e - (empty) this block is empty and can be used
 *   d - (data) this block has data value b3b2b1b0
 *   p - () this node points by key 'spec' to node b3b2b1b0
 *</pre>
 * @author alesak@jikos.cz
 */
public class RafTrie {

	protected RandomAccessFile raf;
	protected long top;
	protected int maxDepth = 512;
	protected long blockSize = 8;
	protected long defaultBlocks = 2;
	protected boolean isDebug = false;
	protected String defaultEncoding = "UTF-8";
	
	protected void log (String l) {
		if (! isDebug) return;
		else System.out.println (l);
	}
	
	public RafTrie (File f) throws IOException {
		if (! f.exists()) {
			f.createNewFile();
			raf = new RandomAccessFile(f, "rw");
			top = 0;
			allocBlocks(defaultBlocks);
			return;
		}
		raf = new RandomAccessFile(f, "rw");

		top = raf.length();
	}
	
	/**
	 * Allocate [count] new blocks at the end of the file.
	 * 
	 * @param count
	 * @return
	 * @throws IOException
	 */
	protected long allocBlocks (long count) throws IOException {
		long pos, np;
		long size = count * blockSize;
		byte b[] = new byte[(int)blockSize];
		pos = top;
		top += size;
		log ("allocBlocks (" + pos + "-" + size + ")");
		raf.setLength(top);
		
		np = pos;
		for (int i = 0; i < count; i ++) {
			if (i == (count-1)) b [0] = 'z';
			else b[0] = 'e';
			
			writeBlock(np, b);
			np += blockSize;
			
		}
		return (pos);
	}
	
	/**
	 * Write block to position "from"
	 * @param from position in file to write 
	 * @param b block
	 * @throws IOException
	 */
	protected void writeBlock (long from, byte []b) throws IOException
	{
		log ("writeBlock: " + from + "," + b.length);
		raf.seek(from);
		raf.write(b, 0, (int)blockSize);
	}
	protected void readBlock (long from, byte [] out) throws IOException
	{
		log ("readBlock:" + from);
		raf.seek(from);
		raf.read(out, 0, (int)blockSize);
	}
	
	/**
	 * Decode long.
	 * 
	 * @param b buffer
	 * @param pos position in buffer
	 * @return decoded long
	 */
	protected long getLong (byte [] b, int pos)
	{
		return ((int)b [0 + pos] & 0xFF) | ((int)b [1 + pos] & 0xFF) << 8 | ((int)b [2 + pos] & 0xFF) << 16 | ((int)b [3 + pos] & 0xFF) << 24;  
	}
	
	/**
	 * Encode long
	 * 
	 * @param b buffer
	 * @param pos position in buffer
	 * @param val value to encode
	 */
	protected void setLong (byte [] b, int pos, long val)
	{
		b [0 + pos] = (byte)((val >> 0) & 0xFF);
		b [1 + pos] = (byte)((val >> 8) & 0xFF);
		b [2 + pos] = (byte)((val >> 16) & 0xFF);
		b [3 + pos] = (byte)((val >> 24) & 0xFF);
	}
	
	/**
	 * Start at position from and find block of specified type
	 * @param from where to start
	 * @param type type (tag) of block to find
	 * @param b buffer to use
	 * @return positiove value of found block, negative value of last block otherise
	 * @throws IOException
	 */
	protected long findBlock (long from, byte type, byte []b) throws IOException
	{
		long p;
		
		p = from;
		
		do {
			readBlock (p, b);
			
			if (b[0] == type) return (p);
			
			else if (b[0] == 'z') return (-1);
			
			else if (b[0] == 'n') p = getLong (b, 2);
			
			else p += blockSize;
		} while (true);
	}
	
	/**
	 * Find trie node pointing at char ch
	 * @param from pos of block to start
	 * @param ch character to find
	 * @param b buffer to use
	 * @return positive if found
	 * @throws IOException
	 */
	protected long findCharBlock (long from, byte ch, byte []b) throws IOException 
	{

		long p = from;
		long np;
		
		do {
			np = findBlock (p, (byte)'p', b);
			
			if (np == -1) {
				return (- findBlock (p, (byte)'z', b));
			}
			
			if (b [1] == ch) return (p);
			
			else p = np + blockSize;
		} while (true);
		
	}

	/**
	 * Find first empty block from specified position, or allocate new.
	 * @param from position to start
	 * @return position of newly allocated block
	 * @throws IOException
	 */
	protected long getEmptyBlock (long from) throws IOException 
	{
		byte b[] = new byte[(int)blockSize];

		long p = from;
		long np;
		long na;
		
		np = findBlock (p, (byte)'e', b);
			
		if (np != -1) return np;
		
		np = findBlock (p, (byte)'z', b);
		
		na = allocBlocks(defaultBlocks);
		
		setLong(b, 2, na);
		b [0] = (byte)'n';
		writeBlock (np, b);
		
		return (na);
	
	}
	
	public interface Callback {
		boolean call(long objId);
	}
	/**
	 * Find objects matching specified value in trie.
	 * 
	 * @param str key to find
	 * @param callback callback called when object (data) is found
	 */
	public void findObject (String str, Callback callback) {
		byte b[] = new byte[(int)blockSize];
		int i;
		long p, np;

		try {
			byte []s = str.getBytes(defaultEncoding);
			p = 0;
			
			for (i = 0; i < s.length; i ++) {
				np = findCharBlock (p, s[i], b);
				if (b [0] != 'p') {
					break;
				} else {
					np = getLong(b, 2);
				}
				p = np;
			}
			
			if (i < s.length) {
				return;
			}
			
			while ((np = findBlock(p, (byte)'d', b)) > 0) {
				Long value = new Long(getLong(b, 2));
				if (callback.call(value)) return;
				p = np + blockSize;
			}
						
		} catch (UnsupportedEncodingException e) {
			throw new Storage.StorageException("Encoding error.");
		} catch (IOException e) {
			throw new Storage.StorageException("Cannot read");
		}
	}

	/**
	 * Finds all objects with such prefix.
	 * 
	 * @param str prefix to find
	 * @param callback callback called when object (data) is found
	 */
	public void findObjectByPrefix (String str, Callback callback) {
		byte b[] = new byte[(int)blockSize];
		int i, lvl;
		long p, np;

		try {
			byte []s = str.getBytes(defaultEncoding);
			p = 0;
			
			for (i = 0; i < s.length; i ++) {
				np = findCharBlock (p, s[i], b);
				if (b [0] != 'p') {
					break;
				} else {
					np = getLong(b, 2);
				}
				p = np;
			}
			
			if (i < s.length) return;
		
			/*
			 * ensure depth-first search of trie tree.
			 * calling callback when leaving node
			 */
			long level[] = new long[maxDepth];
			for (i = 0; i < maxDepth; i ++) level [i] = 0;
			level [0] = p;
			lvl = 0;
			
			do {
				np = findBlock (level [lvl], (byte)'p', b);
				while (b [0] == 'p') {
					byte []bb = new byte[1];
					bb[0] = b [1];
					level [lvl] = np;
					level [++lvl] = getLong (b, 2);
					np = findBlock (level [lvl], (byte)'p', b);
				}
				
				p = level [lvl];
				while ((np = findBlock(p, (byte)'d', b)) > 0) {
					Long value = new Long(getLong(b, 2));
					if (callback.call(value)) return;
					p = np + blockSize;
				}
				lvl --;
				if (lvl >= 0)
				level [lvl] += blockSize;
			} while (lvl >= 0);
						
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
			throw new Storage.StorageException("Encoding error");
		} catch (IOException e) {
			e.printStackTrace();
			throw new Storage.StorageException("Cannot read");
		}
		
		
	}

	public void close ()
	{
		try {
			raf.close();
		} catch (IOException e) {
			throw new Storage.StorageException("Close error");
		} finally {
			raf = null;
		}
	}
	/**
	 * Put object into trie.
	 * 
	 * @param str key
	 * @param objp value
	 */
	protected void putObject (String str, long objp)
	{
		byte b[] = new byte[(int)blockSize];
		int i;
		long p, np, nnp;

		try {
			byte []s = str.getBytes(defaultEncoding);
			p = 0;
			
			for (i = 0; i < s.length; i ++) {
				np = findCharBlock (p, s[i], b);
				//if (np < 0) {
				if (b [0] != 'p') {
					//np = getEmptyBlock (p);
					break;
				} else {
					np = getLong(b, 2);
				}
				p = np;
			}
			
			for (;i < s.length; i ++) {
				nnp = allocBlocks(defaultBlocks);
				np = getEmptyBlock (p);
				b [0] = (byte)'p';
				b [1] = s [i];
				setLong (b, 2, nnp);
				writeBlock(np, b);
				p = nnp;
			}
			
			/*
			 * Check if object with that id is 
			 * not already in list
			 */
			long chnp, chp = p;
			while ((chnp = findBlock(chp, (byte)'d', b)) > 0) {
				/*
				 * don't add it for second time
				 */
				if (getLong(b, 2) == objp) return;
				chp = chnp + blockSize;
			}

			nnp = getEmptyBlock (p);
			b [0] = (byte)'d';
			setLong (b, 2, objp);
			writeBlock (nnp, b);
			
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
			throw new Storage.StorageException("Encoding error");
		} catch (IOException e) {
			e.printStackTrace();
			throw new Storage.StorageException("Cannot read");
		}
	}
}
